@@ -61,7 +61,7 @@
   "Queues."
   :group 'listen)
 
-(defcustom listen-queue-max-probe-processes 16
+(defcustom listen-queue-max-probe-processes (max 1 (/ (num-processors) 2))
   "Maximum number of processes to run while probing track durations."
   :type 'natnum)
 
@@ -91,114 +91,117 @@ intended to be set from the `listen-menu'."
 (declare-function listen-jump "listen")
 (declare-function listen-menu "listen")
 (declare-function listen-pause "listen")
+
+(define-derived-mode listen-queue-mode special-mode "Listen-Queue"
+  (toggle-truncate-lines 1)
+  (hl-line-mode 1)
+  (setq-local bookmark-make-record-function #'listen-queue--bookmark-make-record))
+
 ;;;###autoload
 (defun listen-queue (queue)
   "Show listen QUEUE."
   (interactive (list (listen-queue-complete)))
-  (let* ((buffer-name (format "*Listen Queue: %s*" (listen-queue-name queue)))
-         (buffer (get-buffer buffer-name)))
-    (unless buffer
-      (with-current-buffer (setf buffer (get-buffer-create buffer-name))
-        (let ((inhibit-read-only t))
-          (setf listen-queue queue)
-          (read-only-mode)
-          (erase-buffer)
-          (toggle-truncate-lines 1)
-          (setq-local bookmark-make-record-function #'listen-queue--bookmark-make-record)
-          (when (listen-queue-tracks listen-queue)
-            (make-vtable
-             :columns
-             (list (list :name "▶" :primary 'descend
-                         :getter
-                         (lambda (track _table)
-                           ;; We compare filenames in case the queue's files
-                           ;; have been refreshed from disk, in which case
-                           ;; the track objects would no longer be `eq'.
-                           (if (equal (listen-track-filename track)
-                                      (listen-track-filename (listen-queue-current queue)))
-                               (progn
-                                 (unless (eq (listen-queue-current listen-queue) track)
-                                   ;; HACK: Update current track in queue.  I don't know a
-                                   ;; more optimal place to do this.
-                                   (setf (seq-elt (listen-queue-tracks listen-queue)
-                                                  (seq-position (listen-queue-tracks listen-queue)
-                                                                (listen-queue-current listen-queue)))
-                                         track
-                                         (listen-queue-current listen-queue) track))
-                                 "▶")
-                             " ")))
-                   (list :name "#" :primary 'descend
-                         :getter (lambda (track _table)
-                                   (cl-position track (listen-queue-tracks queue))))
-                   (list :name "Duration"
-                         :getter (lambda (track _table)
-                                   (when-let ((duration (listen-track-duration track)))
-                                     (listen-format-seconds duration))))
-                   (list :name "r/5"
-                         :getter (lambda (track _table)
-                                   (if-let ((rating (map-elt (listen-track-etc track) "fmps_rating"))
-                                            ((not (equal "-1" rating))))
-                                       (progn
-                                         (setf rating (number-to-string (* 5 (string-to-number rating))))
-                                         (propertize rating 'face 'listen-rating))
-                                     "")))
-                   (list :name "Artist" :max-width 20 :align 'right
-                         :getter (lambda (track _table)
-                                   (propertize (or (listen-track-artist track) "")
-                                               'face 'listen-artist)))
-                   (list :name "Title" :max-width 35
-                         :getter (lambda (track _table)
-                                   (propertize (or (listen-track-title track) "")
-                                               'face 'listen-title)))
-                   (list :name "Album" :max-width 30
-                         :getter (lambda (track _table)
-                                   (propertize (or (listen-track-album track) "")
-                                               'face 'listen-album)))
-                   (list :name "#" :align 'right
-                         :getter (lambda (track _table)
-                                   (or (listen-track-number track) "")))
-                   (list :name "Date"
-                         :getter (lambda (track _table)
-                                   (or (map-elt (listen-track-etc track) "originalyear")
-                                       (map-elt (listen-track-etc track) "originaldate")
-                                       (listen-track-date track)
-                                       "")))
-                   (list :name "Genre"
-                         :getter (lambda (track _table)
-                                   (propertize (or (listen-track-genre track) "")
-                                               'face 'listen-genre)))
-                   (list :name "File"
-                         :getter (lambda (track _table)
-                                   (propertize (listen-track-filename track)
-                                               'face 'listen-filename))))
-             :objects-function (lambda ()
-                                 (or (listen-queue-tracks listen-queue)
-                                     (list (make-listen-track :artist "[Empty queue]"))))
-             :sort-by '((1 . ascend))
-             ;; TODO: Add a transient to show these bindings when pressing "?".
-             :actions (list "q" (lambda (_) (bury-buffer))
-                            "?" (lambda (_) (call-interactively #'listen-menu))
-                            "g" (lambda (_) (call-interactively #'listen-queue-revert))
-                            "j" #'listen-jump
-                            "n" (lambda (_) (forward-line 1))
-                            "p" (lambda (_) (forward-line -1))
-                            "m" #'listen-view-track
-                            "N" (lambda (track) (listen-queue-transpose-forward track queue))
-                            "P" (lambda (track) (listen-queue-transpose-backward track queue))
-                            "C-k" (lambda (track) (listen-queue-kill-track track queue))
-                            "C-y" (lambda (_) (call-interactively #'listen-queue-yank))
-                            "RET" (lambda (track) (listen-queue-play queue track))
-                            "SPC" (lambda (_) (call-interactively #'listen-pause))
-                            "o" (lambda (_) (call-interactively #'listen-queue-order-by))
-                            "s" (lambda (_) (listen-queue-shuffle listen-queue))
-                            "l" (lambda (_) "Show (selected) tracks in library view."
-                                  (call-interactively #'listen-library-from-queue))
-                            "!" (lambda (_) (call-interactively #'listen-queue-shell-command)))))
-          (goto-char (point-min))
-          (listen-queue--annotate-buffer)
-          (listen-queue-goto-current)
-          (hl-line-mode 1))))
-    (pop-to-buffer buffer)))
+  (with-current-buffer (get-buffer-create (format "*Listen Queue: %s*" (listen-queue-name queue)))
+    (let ((inhibit-read-only t))
+      (listen-queue-mode)
+      (setf listen-queue queue)
+      (erase-buffer)
+      (when (listen-queue-tracks listen-queue)
+        (make-vtable
+         :columns
+         (list (list :name "▶" :primary 'descend
+                     :getter
+                     (lambda (track _table)
+                       ;; We compare filenames in case the queue's files
+                       ;; have been refreshed from disk, in which case
+                       ;; the track objects would no longer be `eq'.
+                       (if-let ((current-track (listen-queue-current queue))
+                                ((equal (listen-track-filename track)
+                                        (listen-track-filename current-track))))
+                           (progn
+                             (unless (eq track (listen-queue-current listen-queue))
+                               (if-let ((position (seq-position (listen-queue-tracks listen-queue)
+                                                                (listen-queue-current listen-queue))))
+                                   ;; HACK: Update current track in queue.  I don't know a more
+                                   ;; optimal place to do this.
+                                   ;; TODO: Potentially use `listen-queue-track-revert' in more
+                                   ;; places to make this unnecessary.
+                                   (setf (seq-elt (listen-queue-tracks listen-queue) position) track)
+                                 ;; Old track not found: just add it.
+                                 (push track (listen-queue-tracks queue)))
+                               (setf (listen-queue-current listen-queue) track))
+                             "▶")
+                         " ")))
+               (list :name "#" :primary 'descend
+                     :getter (lambda (track _table)
+                               (cl-position track (listen-queue-tracks queue))))
+               (list :name "Duration"
+                     :getter (lambda (track _table)
+                               (when-let ((duration (listen-track-duration track)))
+                                 (listen-format-seconds duration))))
+               (list :name "r/5"
+                     :getter (lambda (track _table)
+                               (if-let ((rating (map-elt (listen-track-etc track) "fmps_rating"))
+                                        ((not (equal "-1" rating))))
+                                   (progn
+                                     (setf rating (number-to-string (* 5 (string-to-number rating))))
+                                     (propertize rating 'face 'listen-rating))
+                                 "")))
+               (list :name "Artist" :max-width 20 :align 'right
+                     :getter (lambda (track _table)
+                               (propertize (or (listen-track-artist track) "")
+                                           'face 'listen-artist)))
+               (list :name "Title" :max-width 35
+                     :getter (lambda (track _table)
+                               (propertize (or (listen-track-title track) "")
+                                           'face 'listen-title)))
+               (list :name "Album" :max-width 30
+                     :getter (lambda (track _table)
+                               (propertize (or (listen-track-album track) "")
+                                           'face 'listen-album)))
+               (list :name "#" :align 'right
+                     :getter (lambda (track _table)
+                               (or (listen-track-number track) "")))
+               (list :name "Date"
+                     :getter (lambda (track _table)
+                               (or (map-elt (listen-track-etc track) "originalyear")
+                                   (map-elt (listen-track-etc track) "originaldate")
+                                   (listen-track-date track)
+                                   "")))
+               (list :name "Genre"
+                     :getter (lambda (track _table)
+                               (propertize (or (listen-track-genre track) "")
+                                           'face 'listen-genre)))
+               (list :name "File"
+                     :getter (lambda (track _table)
+                               (propertize (listen-track-filename track)
+                                           'face 'listen-filename))))
+         :objects-function (lambda ()
+                             (or (listen-queue-tracks listen-queue)
+                                 (list (make-listen-track :artist "[Empty queue]"))))
+         :sort-by '((1 . ascend))
+         ;; TODO: Add a transient to show these bindings when pressing "?".
+         :actions (list "q" (lambda (_) (bury-buffer))
+                        "?" (lambda (_) (call-interactively #'listen-menu))
+                        "g" (lambda (_) (call-interactively #'listen-queue-revert))
+                        "j" #'listen-jump
+                        "n" (lambda (_) (forward-line 1))
+                        "p" (lambda (_) (forward-line -1))
+                        "m" #'listen-view-track
+                        "N" (lambda (track) (listen-queue-transpose-forward track queue))
+                        "P" (lambda (track) (listen-queue-transpose-backward track queue))
+                        "C-k" (lambda (track) (listen-queue-kill-track track queue))
+                        "C-y" (lambda (_) (call-interactively #'listen-queue-yank))
+                        "RET" (lambda (track) (listen-queue-play queue track))
+                        "SPC" (lambda (_) (call-interactively #'listen-pause))
+                        "o" (lambda (_) (call-interactively #'listen-queue-order-by))
+                        "s" (lambda (_) (listen-queue-shuffle listen-queue))
+                        "l" (lambda (_) "Show (selected) tracks in library view."
+                              (call-interactively #'listen-library-from-queue))
+                        "!" (lambda (_) (call-interactively #'listen-queue-shell-command)))))
+      (listen-queue--annotate-buffer)
+      (listen-queue-goto-current))
+    (pop-to-buffer (current-buffer))))
 
 (defun listen-queue--annotate-buffer ()
   "Annotate current buffer.
@@ -323,11 +326,11 @@ return a new one having it.  PROMPT is passed to `format-prompt',
 which see."
   (cl-labels ((read-queue ()
                 (let* ((player (listen-current-player))
-                       (default-queue-name (or (when listen-queue
-                                                 ;; In a listen buffer: offer its queue as default.
-                                                 (listen-queue-name listen-queue))
-                                               (when (listen--playing-p player)
-                                                 (listen-queue-name (map-elt (listen-player-etc player) :queue)))))
+                       (default-queue-name
+                        (when-let ((queue (or listen-queue
+                                              (when (listen--playing-p player)
+                                                (map-elt (listen-player-etc player) :queue)))))
+                          (listen-queue-name queue)))
                        (queue-names (mapcar #'listen-queue-name listen-queues))
                        (prompt (format-prompt prompt default-queue-name))
                        (selected (completing-read prompt queue-names nil (not allow-new-p)
@@ -372,7 +375,6 @@ which see."
            queue)))
   (cl-callf append (listen-queue-tracks queue) (listen-queue-tracks-for files))
   (listen-queue queue)
-  (listen-queue-play queue)
   queue)
 
 (cl-defun listen-queue-add-from-playlist-file (filename queue)
@@ -404,14 +406,26 @@ buffer, if any)."
                      ;; In a queue buffer and the region is active: use it.
                      listen-queue
                    (listen-queue-complete :allow-new-p t)))
-          (tracks (or (if-let ((buffer (listen-queue-buffer queue)))
-                          (with-current-buffer buffer
-                            (when (region-active-p)
-                              (listen-queue-selected))))
-                      (listen-queue-tracks queue))))
-     (list :tracks tracks)))
-  (let ((tracks (or tracks (listen-queue-tracks queue))))
-    (listen-library (mapcar #'listen-track-filename tracks))))
+          (tracks (when-let ((buffer (listen-queue-buffer queue)))
+                    (with-current-buffer buffer
+                      (when (region-active-p)
+                        (listen-queue-selected))))))
+     (list :tracks tracks :queue queue)))
+  (listen-library (or tracks
+                      (lambda ()
+                        (mapcar #'listen-track-filename
+                                ;; In case the queue gets renamed, or gets replaced by a
+                                ;; different one with the same name:
+                                (listen-queue-tracks
+                                 (or (when (member queue listen-queues)
+                                       ;; Ensure the queue is in the queue list (one from a bookmark
+                                       ;; record wouldn't be the same object anymore).  This allows
+                                       ;; a queue to be renamed during a session and still match
+                                       ;; here.
+                                       queue)
+                                     (cl-find (listen-queue-name queue) listen-queues
+                                              :key #'listen-queue-name :test #'equal)
+                                     (error "Queue not found: %S" queue)))))) ))
 
 (defun listen-queue-track (filename)
   "Return track for FILENAME."
@@ -741,8 +755,7 @@ MAX-PROCESSES limits the number of parallel probing processes."
       (with-timeout ((* 0.1 (length tracks)) (error "Probing for track duration timed out"))
         (while (or tracks processes)
           (probe-more)
-          (while (accept-process-output nil 0.01))
-          (sleep-for 0.01))))))
+          (while (accept-process-output nil 0.01)))))))
 
 ;;;;; Queue delay mode
 
